import io
from zenaura.client.page import Page 
from zenaura.client.hydrator import HydratorCompilerAdapter
from zenaura.client.app import App 

compiler_adapter = HydratorCompilerAdapter()

# create pyscript pydido template 
template = lambda content, meta_description=None, title=None, icon=None, pydide="https://pyscript.net/releases/2024.1.1/core.js"  : f"""

<html lang="en">
  <head>
    <meta charset="utf-8" />
    <link rel="icon" href="{icon}" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <meta name="theme-color" content="#000000" />
    <meta name="title" content="{title}" />
    <meta http-equiv="refresh"  />
    <meta
      name="description"
      content="{meta_description}"
    />
    <script type="module" src="{pydide}"></script>

 
	<script type="py" src="./public/main.py" config="./public/config.json"></script>

    <link  rel="stylesheet" href="./public/main.css">

    <title>{title}</title>
    
  </head>
  <body>
    <div id="root">
        {content}
    </div>
  
  </body>

</html>
"""
class ZenauraServer:

    @staticmethod
    def hydrate_page(page : Page, title="zenaura", meta_description="this app created with zenaura", icon="./public/favicon.ico", pydide="https://pyscript.net/releases/2024.1.1/core.js"):
        """
          hyderate zenaura page for server side rendering, 
          params :
           page - zenaura page.
           title - html meta tag for title text
           meta_description : html meta description for website
           icon: favorite.ico
           pydide : is pydide build script url.
        """
        return template(compiler_adapter.hyd_comp_compile_page(page),meta_description, title, icon, pydide)
    
    @staticmethod
    def hydrate_app(app :App, title="zenaura", meta_description="this app created with zenaura", icon="./public/favicon.ico", pydide="https://pyscript.net/releases/2024.1.1/core.js"):
      """
          render pages on app run, set page with path / to visible, rest to hidden
          then compile index.html on server run. 
          params :
          app - zenaura app.
          title - html meta tag for title text
          meta_description : html meta description for website
          icon: favorite.ico
          pydide : is pydide build script url.
      """
      pages = io.StringIO()

      # render pages 
      for path, route in app.routes.items():
          page, _, path, ssr = route 
          if ssr: # ignore SSR pages
              continue 
          if path == "/" : # set / route to visible 
              page_div = lambda comps : f'<div data-zenaura="{page.id}">{comps}</div>'
              pages.write(page_div(compiler_adapter.hyd_comp_compile_page(page)))
              continue
          # pages other than / are set to hidden
          page_div = lambda comps : f'<div hidden data-zenaura="{page.id}">{comps}</div>'
          pages.write(page_div(compiler_adapter.hyd_comp_compile_page(page)))
      

      pages = pages.getvalue() 

      # overwrite in public dir
      with open("./public/index.html", "w") as file:
          print(file)
          file.write(template(pages,meta_description, title, icon, pydide))

        



class Attribute:
    def __init__(self, key, value):
        """
        Initializes an Attribute object with the given key and value.

        Args:
        key: The key of the attribute.
        value: The value of the attribute.
        """
        self.key = key
        self.value = value

    def to_dict(self) -> dict:
        return {
            "key": self.key,
            "value": self.value
        }


from zenaura.client.tags import Node
from typing import List
from zenaura.client.hydrator.compiler_adapter import HydratorCompilerAdapter
from zenaura.client.tags import Data
from itertools import zip_longest
from .operations import *
class Searcher(
    HydratorCompilerAdapter
):
    """
        searching step of zenaura virtual dom algorithm.
    """
    def updater_context_builder(self, name: str, context: dict) -> dict:
        return {
            "name": name,
            "context": context
        }
    def patches_builder(self, prev_child_node, new_child_node, id, child_id):
        return [
                self.hyd_comp_get_keyed_uuid(
                    id=id, 
                    child_id=child_id
                ),
                new_child_node, 
                child_id,
                self.updater_context_builder(
                    name=REMOVE_NODE,
                    context={"children" : prev_child_node}
                )
            ]
    def search(self, prevNode: Node, newNode: Node, id: str) -> List[List[any]]:
        # error handling :
        if not prevNode and not newNode:
            return []
        if not isinstance(prevNode, Node) or not isinstance(newNode, Node):
            return []

        differences = []
        def helper(prev_child_node: Node, new_child_node: Node, id, prev_child_path: str, new_child_path) -> None:
            nonlocal differences
            if not prev_child_node and new_child_node:  # Added
                differences.append([
                self.hyd_comp_get_keyed_uuid(
                            id=id, 
                            key=prev_child_path
                    ),
                    new_child_node, 
                    prev_child_path,
                    self.updater_context_builder(
                        name=ADD_NODE,
                        context={"children" : new_child_node}
                    )
                ])
                return

            if prev_child_node and not new_child_node:  # Removed
                differences.append([
                    self.hyd_comp_get_keyed_uuid(
                        id=id, 
                        key=prev_child_path
                    ),
                    prev_child_node, 
                    prev_child_path,
                    self.updater_context_builder(
                        name=REMOVE_NODE,
                        context={"children" : prev_child_node}
                    )
                ])
                return
            
            if prev_child_node.name != new_child_node.name:  # Changed child by name
                differences.append([
                    self.hyd_comp_get_keyed_uuid(
                        id=id, 
                        key=prev_child_path
                    ),
                    prev_child_node, 
                    prev_child_path,
                    self.updater_context_builder(
                        name=REMOVE_NODE,
                        context={"children" : prev_child_node}
                    )
                ])
                differences.append([
                    self.hyd_comp_get_keyed_uuid(
                        id=id, 
                        key=new_child_path
                    ),
                    new_child_node, 
                    new_child_path,
                    self.updater_context_builder(
                        name=ADD_NODE,
                        context={"children" : new_child_node}
                    )
                ])
                return

            # Compare attributes
            for [prev_attr, new_attr] in zip_longest(prev_child_node.attributes, new_child_node.attributes):
                
                # removed attribute
                if prev_attr and not new_attr:
                    differences.append([
                        self.hyd_comp_get_keyed_uuid(
                            id=id, 
                            key=prev_child_path
                        ),
                        new_child_node, 
                        prev_child_path,
                        self.updater_context_builder(
                            name=REMOVE_ATTRIBUTE,
                            context={"attr_name" : prev_attr.key}
                        )
                    ])
                    continue
                
                # added attribute
                if not prev_attr and new_attr:
                    differences.append([
                        self.hyd_comp_get_keyed_uuid(
                            id=id, 
                            key=new_child_path
                        ), 
                        new_child_node, 
                        new_child_path, 
                        self.updater_context_builder(
                            name=ADD_ATTRIBUTE,
                            context={"attr_name" : new_attr.key, "attr_value" : new_attr.value}
                        )
                    ])
                    continue

                # replaced value attr
                if prev_attr and new_attr:
                    if prev_attr.value != new_attr.value:
                        differences.append([
                           self.hyd_comp_get_keyed_uuid(
                                id=id, 
                                key=prev_child_path
                            ),
                            new_child_node, 
                            prev_child_path,
                            self.updater_context_builder(
                                name=ADD_ATTRIBUTE,
                                context={"attr_name" : new_attr.key, "attr_value" : new_attr.value}
                            )
                        ])



            # Compare children
            for idx, (prev_child, new_child) in enumerate(zip_longest(prev_child_node.children, new_child_node.children)):
                # leaf text nodes
                if prev_child and new_child:
                    if prev_child.is_text_node and new_child.is_text_node and (prev_child.text != new_child.text):
                        differences.append([
                            self.hyd_comp_get_keyed_uuid(
                                id=id, 
                                key=prev_child_path
                            ),
                            new_child_node,
                            prev_child_path,
                            self.updater_context_builder(
                                name=NODE_INNER_TEXT,
                                context={"text" : new_child.text}
                            )
                        ])
                        continue
                new_child_path = new_child.path if isinstance(new_child, Node) else ""
                prev_child_path = prev_child.path if isinstance(prev_child, Node) else ""
                helper(prev_child, new_child, id, prev_child_path=prev_child_path, new_child_path=new_child_path)

        helper(prevNode, newNode, id, 0, 0)
        return differences


import html
import bleach

from zenaura.client.config import (
    allowed_tags,
    allowed_attributes,
)

class CompilerSanitizer:
    def sanitize(
                self, 
                user_input, 
                allowed_tags=allowed_tags, 
                allowed_attributes=allowed_attributes
            ):
        """
        Sanitizes user input to prevent various injection attacks.

        Args:
            user_input (str): The raw user input to sanitize.
            allowed_tags (list): A list of allowed HTML tags (e.g., ['p', 'br', 'strong']).
            allowed_attributes (dict): A dictionary mapping allowed tags to their allowed 
                                    attributes (e.g., {'img': ['src', 'alt']}).  

        Returns:
            str: The sanitized input.
        """
        user_input = str(user_input)
        
        safe_html = html.escape(user_input)  # Escape all HTML special characters initially
        safe_html = bleach.clean(
            safe_html, tags=allowed_tags, attributes=allowed_attributes
            ) 
      

        return safe_html 


ZENAURA_DOM_ATTRIBUTE = "data-zenaura"

# TODO : add more
# TODO add more 
allowed_tags = [
    'p', 'br', 'strong', 'em', 'b', 'i', 'u', 's', 'del', 'ins', 'sub', 'sup', 
    'ul', 'ol', 'li',
    'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
    'a', 'img',
    'table', 'thead', 'tbody', 'tr', 'th', 'td',
    'blockquote', 'q',
    'pre', 'code'
]

# TODO : add more
allowed_attributes = {
    '*': ['class', 'id', 'title', 'lang', 'dir', 'data-*', 'aria-*', 'role', 'py-*'], # py-* for pyscript 
    'a': ['href', 'target', 'rel'],
    'img': ['src', 'alt', 'width', 'height'],
}

self_closing_tags = [
    'area', 'base', 'br', 'col', 'embed', 'hr', 'img', 'input', 'link',
     'meta', 'param', 'source', 'track', 'wbr'
]



from zenaura.client.dom import zenaura_dom
import functools

def mutator(func):
    """
    Decorator function to render the DOM after the execution of the decorated function.

    Args:
    - func: The function to be decorated.

    Returns:
    - wrapper_func: The wrapper function that executes the original function and then renders the DOM.
    """
    @functools.wraps(func)
    async def wrapper_func(self, *args, **kwargs):
        """
        Wrapper function that executes the original function and then renders the DOM.

        Args:
        - self: The instance of the class.
        - *args: Positional arguments for the original function.
        - **kwargs: Keyword arguments for the original function.

        Returns:
        - None
        """
        await func(self, *args, **kwargs)
        await zenaura_dom.render(self)
    return wrapper_func


from .hydrator import Hydrator
from .compiler_adapter import HydratorCompilerAdapter
from .real_dom_adapter import HydratorRealDomAdapter
from .virtual_dom_adapter import HydratorVirtualDomAdapter
from .tasker import HydratorTasker


import uuid 
import itertools
from .component import Component, UUIDManager
from typing import List


class Page:
    """
        a class representing a page in zenaura application
    """
    _page_count = itertools.count(0)

    def __init__(self, children : List[Component]):
        self.count = next(self._page_count)
        self.id = UUIDManager.generate_uuid(self.__class__.__name__, self.count)
        self.children = children
        if not isinstance(self.children, list):
            raise TypeError("children must be a list")

        for child in self.children:
            if not isinstance(child, Component):
                raise TypeError("page children must be a Component")


    
        


from abc import ABC, abstractmethod

class Observer(ABC):
    """
    Create observers for subjects to notify.
    """

    @abstractmethod
    def update(self, value):
        """
        Update method to be implemented by concrete observers.

        Parameters:
        value (dict): The updated value from the subject.

        Returns:
        None
        """
        pass


class Content:
    def __init__(self):
        self.firstChild = ""

class MockTextNode:
    def __init__(self, text):
        self.nodeValue = text
        self.parentNode = None

class MockElement:
    def __init__(self, tag_name=None, innerHTML=""):
        self.tagName = tag_name
        self.innerHTML = innerHTML
        self.outerHTML = ""
        self.attributes = {}
        self.childNodes = []
        self.parentNode = None
        self.content = Content()

    def setAttribute(self, name, value):
        self.attributes[name] = value

    def getAttribute(self, name):
        return self.attributes.get(name)

    def removeAttribute(self, name):
        self.attributes.pop(name, None)

    def appendChild(self, child):
        if child not in self.childNodes:
            self.childNodes.append(child)
            if isinstance(child, MockElement):
                child.parentNode = self

    def removeChild(self, child):
        if child in self.childNodes:
            self.childNodes.remove(child)
            child.parentNode = None

    def createElement(self, tag_name):
        return MockElement(tag_name)
    

    def __repr__(self):  # Helpful for debugging
        return f"<MockElement '{self.tagName}'>"

class MockDocument:  # Use MagicMock for flexibility
    def __init__(self):
        self.body = MockElement("body")  # Create a body element
        self.elementsById = {"root": self.body}  # Store elements by ID
        self.title = ""

    def getElementById(self, element_id):
        return self.elementsById.get(element_id)

    def createElement(self, tag_name):
        return MockElement(tag_name)
    
    def setElementById(self, element_id, element):
        # for mocking purposes
        self.elementsById[element_id] = element

    def createTextNode(self,txt):
        textNode = MockTextNode(txt)
        return textNode

    def querySelector(self, query:str):
        query = query.replace("[", "").replace("]", "").replace('"', "").split("=")
        id = query[-1]
        return self.elementsById[id]

class MockLocation:
    def __init__(self):
        self.href = "http://localhost:8000"  # Example
        self.pathname = ""

class MockWindowHistory:
    def __init__(self):
        self.history = []

    def pushState(self, *args, **kwargs):
        self.history.append([args, kwargs])

class MockWindow:
    def __init__(self):
        self.innerWidth = 1024 
        self.innerHeight = 768
        self.location = MockLocation()
        self.history = MockWindowHistory()



from zenaura.client.tags import Node, Data
from .attribute import AttributeProccessor
from .sanitize import CompilerSanitizer
import io
from zenaura.client.config import (
    ZENAURA_DOM_ATTRIBUTE,
    self_closing_tags
)



class Compiler(
    CompilerSanitizer,
    AttributeProccessor,
):
    def __init__(self):
        """
            Constructor for the Compiler class.
            Initializes the attribute keyword mapping.

            Args:
                self: The Compiler instance.

            Returns:
                None
        """
        self.attrKeyWords= {
             "styles" : "class"
        }
    
    def getKeyedUID(self, 
                    id, 
                    withAttribut=False, 
                    key=None
                ):
        """
            from unique UID for component generate unique
            keyed UUID for each child within the component tree.
            This is used to create keyed ZENAURA_DOM_ATTRIBUTE

            args :
                id (str): unique id for the component
                child_id: which is unique counter for the child position within the component
                withAttribute (bool, optional): adds unique attribute for virtual dom.
            used in :
                - compiler
                - mount life cycle 
                - render life cycle
            returns :
                if withAttribute is True :
                {ZENAURA_DOM_ATTRIBUTE}="{id}{child_id}"
                else :
                {id}{child_id}
        """
        if isinstance(key, list):
            key = "".join(str(i) for i in key) # [0,0,0,1] ->  0001
        if withAttribut:
            return f' {ZENAURA_DOM_ATTRIBUTE}="{id}{key}"'
        return f'{id}{key}'
    
    def compile(
        self, 
        elm: Node,
        id=None, 
        zenaura_dom_mode=False
    ):
        """
        Compiles a Zenui Node into its corresponding HTML representation.

        Args:
            elm (Node): The Zenui Node object to compile.
            id (str, optional): used to create keyed ZENAURA_DOM_ATTRIBUTE
            attribute UID-parent-child-child-child on so on.
            zenaura_dom_mode (bool, optional): Adds unique attribute for virtual dom.

        Returns:
            str: A string containing the compiled HTML.
        """
        if elm.is_text_node or isinstance(elm, str):
            return self.sanitize(elm.text)

        tag = elm.name 

        zenui_id = ""
    
        #  assign unique id for zenui dom
        if isinstance(elm, Node) and zenaura_dom_mode:
            zenui_id = self.getKeyedUID(
                id, 
                withAttribut=True, 
                key=elm.path
            )

        # get node attributes
        attributes = self.process_attributes(elm.attributes)
        
        if tag in self_closing_tags:
            return f"<{tag}{zenui_id}{attributes}>"
        
        # start tag
        html = io.StringIO()
        html.write(f"<{tag}{zenui_id}{attributes}>")

        # get children
        for child in elm.children:
            html.write(
                self.compile(
                    child, 
                    id, 
                    zenaura_dom_mode, 
                )
            )

                    
        html.write(f"</{tag}>")

        # finish tag
        return html.getvalue()


from zenaura.client.tags import Node, Attribute, Builder
import unittest

class DataclassTests(unittest.TestCase):


    def test_attribute_creation(self):
        attribute = Attribute(key="test", value="test")
        self.assertEqual(attribute.key, "test")
        self.assertTrue(isinstance(attribute, Attribute))

    def test_node_creation(self):
        child = Node(name="test")
        node = Node(name="div")
        node.append_child(child)
        node.attributes.append(Attribute(key="test", value="test"))
        self.assertEqual(child.name, "test")
        self.assertEqual(len(node.children), 1)
        self.assertEqual(len(node.attributes), 1)
        self.assertTrue(isinstance(node.children[0], Node ))
        self.assertTrue(isinstance(node.attributes[0], Attribute))

    def test_node_has_nodeId(self):
        child = Node(name="test")
        parent = Node(name="div")
        parent.append_child(child)
        self.assertTrue(child.nodeId)
        self.assertTrue(parent.nodeId)

    def test_node_builder(self):
        node = Builder('div') \
            .with_attribute("test", "test") \
            .with_child(
                Node(name="test")
            ).build()
        
        self.assertEqual(len(node.children), 1)
        self.assertEqual(len(node.attributes), 1)
        self.assertTrue(isinstance(node.children[0], Node ))
        self.assertTrue(isinstance(node.attributes[0], Attribute))

    def test_init(self):
        tb = Builder('div').build()
        self.assertEqual(tb.name, "div")
        self.assertEqual(tb.attributes, [])  
        self.assertEqual(tb.children, []) 

    def test_with_attribute(self):
        tb = Builder('p')
        tb.with_attribute("id", "main-paragraph").build()
        self.assertEqual(tb.node.attributes[0].key, "id")
        self.assertEqual(tb.node.attributes[0].value, "main-paragraph")


    def test_with_styles(self):
        tb = Builder('p')
        tb.with_styles({"color": "blue", "font-size": "16px"}).build()
        self.assertEqual(tb.node.attributes[0].key, "style")
        self.assertEqual(tb.node.attributes[0].value, "color:blue;font-size:16px")

    def test_with_class(self):
        tb = Builder('p')
        tb.with_class("my-class").build() 
        self.assertEqual(tb.node.attributes[0].key, "class")
        self.assertEqual(tb.node.attributes[0].value, "my-class")

    def test_with_classes(self):
        tb = Builder('span')
        tb.with_classes("highlighted", "bold").build()
        self.assertEqual(tb.node.attributes[0].key, "class")
        self.assertEqual(tb.node.attributes[0].value, "highlighted bold")


    def test_with_class_avoid_duplicates(self):
        tb = Builder('div')
        tb.with_class("container").with_class("container").build()
        self.assertEqual(tb.node.attributes[0].key, "class")
        self.assertEqual(tb.node.attributes[0].value, "container")

    def test_node_to_html_with_nested_nodes(self):
        # Create a nested node structure
        node1 = Node("div")
        node1.attributes.append(Attribute("class", "container"))

        node2 = Node("h1")
        node2.append_child("Welcome to our website")

        node3 = Node("p")
        node3.append_child("This is a paragraph")

        node1.append_child(node2)
        node1.append_child(node3)

        # Convert the node to HTML
        expected_html = (
            "<div class=\"container\">"
            "<h1>Welcome to our website</h1>"
            "<p>This is a paragraph</p>"
            "</div>"
        )
        self.assertEqual(node1.to_html(), expected_html)

    def test_node_to_html_with_nested_nodes_and_attributes(self):
        # Create a nested node structure with attributes
        node1 = Node("div")
        node1.attributes.append(Attribute("class", "container"))

        node2 = Node("h1")
        node2.attributes.append(Attribute("id", "title"))
        node2.append_child("Welcome to our website")

        node3 = Node("p")
        node3.attributes.append(Attribute("class", "description"))
        node3.append_child("This is a paragraph")

        node1.append_child(node2)
        node1.append_child(node3)

        # Convert the node to HTML
        expected_html = (
            "<div class=\"container\">"
            "<h1 id=\"title\">Welcome to our website</h1>"
            "<p class=\"description\">This is a paragraph</p>"
            "</div>"
        )
        self.assertEqual(node1.to_html(), expected_html)

    def test_node_to_html_with_nested_nodes_and_mixed_content(self):
        # Create a nested node structure with mixed content
        node1 = Node("div")
        node1.attributes.append(Attribute("class", "container"))

        node2 = Node("h1")
        node2.append_child("Welcome to our website")

        node3 = Node("p")
        node3.append_child("This is a paragraph")

        node4 = Node("img")
        node4.attributes.append(Attribute("src", "image.jpg"))
        node4.attributes.append(Attribute("alt", "Image"))

        node1.append_child(node2)
        node1.append_child(node3)
        node1.append_child(node4)

        # Convert the node to HTML
        expected_html = (
            "<div class=\"container\">"
            "<h1>Welcome to our website</h1>"
            "<p>This is a paragraph</p>"
            "<img src=\"image.jpg\" alt=\"Image\">"
            "</div>"
        )
        self.assertEqual(node1.to_html(), expected_html)

    def test_node_to_html_with_nested_nodes_and_empty_children(self):
        # Create a nested node structure with empty children
        node1 = Node("div")
        node1.attributes.append(Attribute("class", "container"))

        node2 = Node("h1")
        node2.append_child("Welcome to our website")

        node3 = Node("p")
        node3.append_child("")

        node1.append_child(node2)
        node1.append_child(node3)

        # Convert the node to HTML
        expected_html = (
            "<div class=\"container\">"
            "<h1>Welcome to our website</h1>"
            "<p></p>"
            "</div>"
        )
        self.assertEqual(node1.to_html(), expected_html)



       
        
    def test_to_dict_without_children(self):
        
        node = Node("parent")

        
        result = node.to_dict()

        

        self.assertDictEqual(
            result, {'name': 'parent', 'parent': 'none', 'level': 0, 'key': 0, 'path': '', 'children': []}
        )
        
    def test_to_dict_with_nested_children(self):
        
        child1 = Node("child1", children=[Node("grandchild1"), Node("grandchild2")])
        child2 = Node("child2")
        node = Node("parent", children=[child1, child2])

        
        result = node.to_dict()
        
        self.assertDictEqual(
            result,
{'name': 'parent', 'parent': 'none', 'level': 0, 'key': 0, 'path': '', 'children': [{'name': 'child1', 'parent': 'child1', 'level': 1, 'key': 0, 'path': '00', 'children': [{'name': 'grandchild1', 'parent': 'grandchild1', 'level': 2, 'key': 0, 'path': '0010', 'children': []}, {'name': 'grandchild2', 'parent': 'grandchild2', 'level': 2, 'key': 1, 'path': '0011', 'children': []}]}, {'name': 'child2', 'parent': 'child2', 'level': 1, 'key': 1, 'path': '01', 'children': []}]}
        )
    def test_getAttributes(self):

        node = Node("div")
        node.attributes.append(Attribute("class", "container"))
        node.attributes.append(Attribute("id", "title"))
        node.attributes.append(Attribute("style", "color:blue;font-size:16px"))

        attributes = node.getAttributes(node)

        expected = " class=\"container\" id=\"title\" style=\"color:blue;font-size:16px\""
        self.assertEqual(attributes, expected)


    def test_findChildByName_found(self):
        node = Node("div")
        node.append_child(Node("h1"))

        child = node.findChildByName("h1")
        self.assertEqual(child.name, "h1")

    def test_findChildByName_not_found(self):
        node = Node("div")

        child = node.findChildByName("h1")
        self.assertFalse(child)

    def test_findAllByName_found(self):
        node = Node("div")
        node.append_child(Node("h1"))
        node.append_child(Node("h1"))
        node.append_child(Node("p"))

        children = node.findAllByName("h1")
        self.assertEqual(len(children), 2)

    def test_findAllByName_not_found(self):
        node = Node("div")
        node.append_child(Node("h1"))
        node.append_child(Node("p"))

        children = node.findAllByName("h2")
        self.assertEqual(len(children), 0)

    def test_findByAttribute_found(self):
        node = Node("div")
        child = Node("h1")
        child.attributes.append(Attribute("class", "container"))
        node.append_child(child)

        child = node.findByAttribute("class", "container")
        self.assertEqual(child.name, "h1")

    def test_findByAttribute_not_found(self):
        node = Node("div")

        child = node.findByAttribute("class", "container")
        self.assertFalse(child)

    def test_findAllChildrenByAttributeKey_found(self):
        node = Node("div")
        node.append_child(Node("h1", attributes=[Attribute("class", "container")]))
        node.append_child(Node("h1", attributes=[Attribute("class", "container")]))
        node.append_child(Node("p"))

        children = node.findAllChildrenByAttributeKey("class")
        self.assertEqual(len(children), 2)

    def test_findAllChildrenByAttributeKey_not_found(self):
        node = Node("div")
        node.append_child(Node("p"))

        children = node.findAllChildrenByAttributeKey("class")
        self.assertFalse(children)

    def test_findAllChildrenByAttributeValue_found(self):
        node = Node("div")
        node.append_child(Node("h1", attributes=[Attribute("class", "container")]))
        node.append_child(Node("h1", attributes=[Attribute("class", "container")]))
        node.append_child(Node("p"))

        children = node.findAllChildrenByAttributeValue("container")
        self.assertEqual(len(children), 2)


    def test_findAllChildrenByAttributeValue_not_found(self):
        node = Node("div")

        children = node.findAllChildrenByAttributeValue("container")
        self.assertFalse(children)

    def test_replace_children_found(self):
        node = Node("div")
        spe = Node("h1")
        node.append_child(Node("h2"))
        node.append_child(spe)
        node.append_child(Node("p"))

        node.replace(spe, Node("h2"))
        self.assertEqual(len(node.children), 3)
        self.assertEqual(node.children[0].name, "h2")
        self.assertEqual(node.children[1].name, "h2")
        self.assertEqual(node.children[2].name, "p")

    def test_replace_children_not_found(self):
        node = Node("div")
        spe = Node("h1")
        node.append_child(Node("h2"))
        node.append_child(Node("h1"))
        node.append_child(Node("p"))

        node.replace(spe, spe)
        self.assertEqual(len(node.children), 3)
        self.assertEqual(node.children[0].name, "h2")
        self.assertEqual(node.children[1].name, "h1")
        self.assertEqual(node.children[2].name, "p")

    def test_getChildIndex_found(self):
        node = Node("div")
        k = Node("h1")
        node.append_child(k)
        node.append_child(Node("h2"))
        node.append_child(Node("p"))

        index = node.getChildIndex(k)
        self.assertEqual(index, 0)

    def test_insertAfter(self):
        node = Node("div")
        h2 = Node("h2")
        node.append_child(Node("h1"))
        node.append_child(h2)
        node.append_child(Node("p"))

        node.insertAfter(h2, Node("h3"))
        self.assertEqual(len(node.children), 4)
        self.assertEqual(node.children[0].name, "h1")
        self.assertEqual(node.children[1].name, "h2")
        self.assertEqual(node.children[2].name, "h3")
        self.assertEqual(node.children[3].name, "p")

    def test_insertBefore(self):
        node = Node("div")
        h2 = Node("h2")
        node.append_child(Node("h1"))
        node.append_child(h2)
        node.append_child(Node("p"))

        node.insertBefore(h2, Node("h3"))
        self.assertEqual(len(node.children), 4)
        self.assertEqual(node.children[0].name, "h1")
        self.assertEqual(node.children[1].name, "h3")
        self.assertEqual(node.children[2].name, "h2")
        self.assertEqual(node.children[3].name, "p")


    def test_remove(self):
        node = Node("div")
        h2 = Node("h2")
        node.append_child(Node("h1"))
        node.append_child(h2)
        node.append_child(Node("p"))

        node.remove(h2)
        self.assertEqual(len(node.children), 2)
        self.assertEqual(node.children[0].name, "h1")
        self.assertEqual(node.children[1].name, "p")

    def test_appendAttributeToChild(self):

        node = Node("div")
        h2 = Node("h2")
        node.append_child(Node("h1"))
        node.append_child(h2)
        node.append_child(Node("p"))

        h2.attributes.append(Attribute("class", "container"))
        h2.attributes.append(Attribute("id", "title"))
        h2.attributes.append(Attribute("style", "color:blue;font-size:16px"))

        node.appendAttributeToChild(h2, Attribute("class", "container2"))

        self.assertEqual(len(h2.attributes), 4)
        self.assertEqual(h2.attributes[0].key, "class")
        self.assertEqual(h2.attributes[0].value, "container")
        self.assertEqual(h2.attributes[1].key, "id")
        self.assertEqual(h2.attributes[1].value, "title")
        self.assertEqual(h2.attributes[2].key, "style")
        self.assertEqual(h2.attributes[2].value, "color:blue;font-size:16px")
        self.assertEqual(h2.attributes[3].key, "class")
        self.assertEqual(h2.attributes[3].value, "container2")
        

    def test_with_attributes_with_children(self):
        # Create a builder object with tag 'div'
        node = Builder('div') \
            .with_attribute("test", "test") \
            .with_children(
                Node(name="test_child1"),
                Node(name="test_child2")
            ) \
            .with_attributes(
                attr1="value1",
                attr2="value2"
            ).build()

        # Assert attributes
        self.assertEqual(node.attributes[0].key, "test")
        self.assertEqual(node.attributes[0].value, "test")
        self.assertEqual(node.attributes[1].key, "attr1")
        self.assertEqual(node.attributes[1].value, "value1")
        self.assertEqual(node.attributes[2].key, "attr2")
        self.assertEqual(node.attributes[2].value, "value2")

        # Assert children
        self.assertEqual(len(node.children), 2)
        self.assertEqual(node.children[0].name, "test_child1")
        self.assertEqual(node.children[1].name, "test_child2")


from zenaura.client.tags import Node
from zenaura.client.page import Page
from zenaura.client.compiler import compiler
from zenaura.client.config import ZENAURA_DOM_ATTRIBUTE
from zenaura.client.component import Component
from typing import List
import io

class HydratorCompilerAdapter:
    """
        hyderator adapter for all real dom operations
        methods should start with:
        hyd_comp_
    """        
    def hyd_comp_get_keyed_uuid(
        self,
        id : str, 
        key : str
    ):
        """
            compiler operation : wraps compiler.getKeyedUID 
        """
        return compiler.getKeyedUID(
            id=id, 
            key=key,
        )
    
    def hyd_comp_compile_render(
        self,
        comp: Component,
    ):
        """
            compiler operation : wraps compiler compile, returns str "HTMLElement"
            compile node
        """
        return compiler.compile(
                comp.render(), 
                id=comp.id,
                zenaura_dom_mode=True
            )
    
    def hyd_comp_compile_children(
        self,
        children: List[Node],
        id: str,
        zenaura_dom_mode: bool,
        key: str=""
    ):
        """
            compiler operation : wraps compiler compile, returns str "HTMLElement"
            compiles node children
        """
        return compiler.compile(
            children, 
            id=id,
            zenaura_dom_mode=True,
        )
    
    def hyd_comp_compile_page(self, page: Page) -> str:
        html = io.StringIO()
        for comp in page.children:
            html.write(
                compiler.compile(
                    comp.render(), 
                    comp.id,
                    zenaura_dom_mode=True,
                )
            )
        return html.getvalue()
    




from .mount import Mount
from .render import Render

class Dom(
    Mount,
    Render
):
    """
        Zenaura virtual dom.
        Mount is responsible for mounting lifecycle of component.
        Render is responsible for rendering lifecycle of component.
    """



from .lifecycles.mount import MountLifeCycles
from .error import GracefulDegenerationLifeCycleWrapper
from zenaura.client.page import Page
from zenaura.client.hydrator import HydratorRealDomAdapter
import traceback

rdom_hyd = HydratorRealDomAdapter()

class Mount(
    GracefulDegenerationLifeCycleWrapper,
    MountLifeCycles,
    ):
    async def mount(self, page: Page) -> None:
        """
            Mount only Page instance to the DOM.
            Only one page instance can be mounted at a time.
            Lifecycle:
                 server on run already hydrate all app pages and overwrite index.html
                 in App class we toggle hidden visibility 
                 in Mount we just trigger attached lifecycle. 
            try : 
                - mount the component.
            except:
                - call componentDidCatchError method.
            Parameters:
            - comp: An instance of the Component class.

            params : 
             page - zenaura page which is list of components with unique ID. 
            Returns:
            None
        """

        try :
            for comp in page.children:
                # trigger attached for page components
                await self.attached(comp)
                # update state in vdom
                self.hyd_vdom_update(comp)

        except Exception as e:
            self.componentDidCatchError(page.children[0], traceback.format_exc())


import sqlite3

class PersistenceRegistry:
    def __init__(self, db_file):
        self.conn = sqlite3.connect(db_file)
        self.create_tables()

    def create_tables(self):
        cursor = self.conn.cursor()

        # Table to map UUIDs to integers
        cursor.execute('''CREATE TABLE IF NOT EXISTS uuid_mapping (
                            uuid TEXT PRIMARY KEY,
                            integer_id INTEGER UNIQUE
                        )''')

        # Table to map HTML components to UUIDs and their content
        cursor.execute('''CREATE TABLE IF NOT EXISTS html_mapping (
                            component_id TEXT PRIMARY KEY,
                            html_content TEXT
                        )''')

        self.conn.commit()
    
    def find_or_create_uuid_integer_mapping(self, uuid_str, integer_id):
        existing_uuid = self.retrieve_uuid(integer_id)
        if existing_uuid:
            return existing_uuid
        return self.insert_uuid(uuid_str, integer_id)

    def insert_uuid(self, uuid_str, integer_id):
        cursor = self.conn.cursor()
        cursor.execute('''INSERT INTO uuid_mapping (uuid, integer_id)
                          VALUES (?, ?)''', (uuid_str, integer_id))
        self.conn.commit()

    def retrieve_uuid(self, integer_id):
        cursor = self.conn.cursor()
        cursor.execute('''SELECT uuid FROM uuid_mapping
                          WHERE integer_id = ?''', (integer_id,))
        row = cursor.fetchone()
        return row[0] if row else None

    def insert_html(self, component_id, html_content):
        cursor = self.conn.cursor()
        cursor.execute('''INSERT INTO html_mapping (component_id, html_content)
                              VALUES (?, ?)''', (component_id, html_content))
        self.conn.commit()

    def retrive_html(self, component_id):
        cursor = self.conn.cursor()
        cursor.execute('''SELECT uuid FROM html_mapping
                          WHERE component_id = ?''', (component_id,))
        row = cursor.fetchall()
        print("row", row)
        return row[0] if row else None
registry = PersistenceRegistry("./persistance.db")


from abc import ABC, abstractmethod

class Subject:
    """
    Create subjects for components to communicate on.
    """

    def __init__(self):
        """
        Initialize a new Subject.

        Parameters:
        None

        Returns:
        None
        """
        self._observers = set()
        self._state = {}

    def attach(self, observer):
        """
        Attach an observer to the subject.

        Parameters:
        observer (Observer): The observer to attach.

        Returns:
        None
        """
        self._observers.add(observer)

    def detach(self, observer):
        """
        Detach an observer from the subject.

        Parameters:
        observer (Observer): The observer to detach.

        Returns:
        None
        """
        self._observers.discard(observer)

    def notify(self):
        """
        Notify all attached observers.

        Parameters:
        None

        Returns:
        None
        """
        for observer in self._observers:
            observer.update(self._state)

    @property
    def state(self):
        """
        Get the state of the subject.

        Parameters:
        None

        Returns:
        dict: The state of the subject.
        """
        return self._state

    @state.setter
    def state(self, new_value):
        """
        Set the state of the subject and notify observers.

        Parameters:
        new_value (dict): The new state of the subject.

        Returns:
        None
        """
        self._state = new_value
        self.notify()


class Observer(ABC):
    """
    Create observers for subjects to notify.
    """

    @abstractmethod
    def update(self, value):
        """
        Update method to be implemented by concrete observers.

        Parameters:
        value (dict): The updated value from the subject.

        Returns:
        None
        """
        pass


from collections import defaultdict
import asyncio
class HydratorTasker:
    """
        manage tasks coming from zenaura diffing algorithm:
            for:
                1. updating real dom.
                2. updating virtual dom.
            do: 
             1. create task queue if not created for component id.
             2. return task queue by component id.
            task queue:
             3. enqueue tasks for components.
             4. dequeue tasks for components.

        component-id : asyncio-task-queue
    """
    queue_lookup = defaultdict(lambda : None)

    def __init__(self):
        pass

    def hyd_tsk_get_or_create_task_queue(self, component_id) -> asyncio.Queue:
        """
            gets or create task queue for component_id
            args:
                component_id: str
            returns:
                task queue
        """
        try :
            if component_id not in self.queue_lookup:
                self.queue_lookup[component_id] = asyncio.Queue()
                return self.queue_lookup[component_id]
            else:
                return self.queue_lookup[component_id]
        except KeyError:
            return False

    
    def hyd_tsk_enqueue_task(self, component_id, task):
        """
            enque task for component if not full and exists
            returns True if task is enqueued else False
        """
        comp_queue = self.queue_lookup[component_id]
        try :
            comp_queue.put_nowait(task)
            return True
        except asyncio.QueueFull:
            return False

    async def hyd_tsk_do_nothing(self):
        """
            final call to queue tasks when que is empty
        """
        pass
    
    def hyd_tsk_dequeue_task(self, component_id):
        """
            dequeue task for component if exists
            else return False        
        """
        comp_queue = self.queue_lookup[component_id]
        try :
            task = comp_queue.get_nowait()
            return task
        except asyncio.QueueEmpty:
            # clean up and return empty function
            del self.queue_lookup[component_id]
            return self.hyd_tsk_do_nothing


from .attribute import Attribute
from .node import Node

class Builder:
    def __init__(self, name : str ="div") -> None:
        """
        Initializes a Builder object with the given name.

        Args:
        name (str): The name of the tag.
        """
        self.node = Node(name)
    
    def with_attributes(self, **kwargs) -> "Builder":
        """
        Adds attributes to the tag.

        Args:
            **kwargs: Key-value pairs of attributes.

        Returns:
            Builder: The Builder object.
        """
        for key, value in kwargs.items():
            self.with_attribute(key, value)
        return self
    
    def with_children(self, *children) -> "Builder":
        """
        Adds child nodes to the tag.

        Args:
            *children: List of child nodes.

        Returns:
            Builder: The Builder object.
        """
        for child in children:
            self.with_child(child)
        return self
    
    def with_attribute(self, key : str, value : any) -> "Builder":
        """
        Adds an attribute to the tag.

        Args:
        key (str): The key of the attribute.
        value: The value of the attribute.

        Returns:
        Builder: The Builder object.
        """
        self.node.attributes.append(Attribute(key,value))
        return self
    
    def with_child(self, child : Node) -> "Builder":
        """
        Adds a child node to the tag.

        Args:
        child (Node): The child node to be added.

        Returns:
        Builder: The Builder object.
        """
        self.node.append_child(child)
        return self 
    
    def with_styles(self, styles: dict) -> "Builder":
        """
        Adds styles to the tag.

        Args:
        styles (dict): Dictionary of styles.

        Returns:
        Builder: The Builder object.
        """
        style_str = ";".join([f"{k}:{v}" for k, v in styles.items()])
        self.node.attributes.append(Attribute("style", style_str))
        return self
    
    def with_classes(self, *class_names: str) -> "Builder":

        """
        Adds multiple class names to the element.

        Args:
        *class_names (str): Variable number of class names.

        Returns:
        Builder: The Builder object.
        """
        for class_name in class_names:
            self.with_class(class_name)
        return self
    
    def with_class(self, class_name: str) -> "Builder":
        """
        Adds a single class name to the element.

        Args:
        class_name (str): The class name to be added.

        Returns:
        Builder: The Builder object.
        """
        for i in self.node.attributes:
            if i.key =="class":
                if class_name not in i.value:
                    i.value = i.value + " " +  class_name
                    return self
        self.node.attributes.append(Attribute("class", class_name))
       
        return self

    def with_class_if(self, class_name: str, condition: bool) -> "Builder":
        """
        Adds a class name to the element if the condition is True.
        If the condition is False, the class is not added.

        args : 
            class_name (str): The class name to be added.
            condition (bool): The condition for adding the class.
        """
        self.with_class(class_name) if condition else None
        return self 

    def with_attribute_if(self, key : str, value : any, condition: bool) -> "Builder":
        """
            adds attribute if condition is true 
            args :
            key (str): The key of the attribute.
            value: The value of the attribute.
            condition (bool): The condition for adding the attribute.
        """
        self.with_attribute(key, value) if condition else None
        return self
    
    def with_child_if(self, child : Node, condition: bool) -> "Builder":
        """
            adds child if condition is true
            args :
            child (Node): The child node to be added.
            condition (bool): The condition for adding the child.
        """
        self.with_child(child) if condition else None
        return self
    
    def with_text(self, text : str):
        """
            add text node
        """
        self.with_child(Node(text=text))
        return self

    
    def build(self):
        """
        Builds and returns the node.

        Returns:
        Node: The built node.
        """
        return self.node


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='zenaura',
    version='0.9.91',
    description="Zenaura is an experimental Python library built upon PyScript, designed to empower Python developers to create stateful, component-based Single Page Applications (SPAs). By leveraging a virtual DOM implementation, Zenaura optimizes the performance, reactivity, responsiveness, and interactivity of web applications. This allows developers to build high-performance, dynamic web applications using familiar Python concepts and syntax.",
    author="Ahmed Rakan",
    author_email="ar.aldhafeeri11@gmail.com",
    packages=['zenaura', 'zenaura.server', 'zenaura.client', 'zenaura.client.algorithm', 'zenaura.client.compiler', 'zenaura.client.hydrator', 'zenaura.client.observer', 'zenaura.client.tags', 'zenaura.client.dom', 'zenaura.client.dom.lifecycles'],
    install_requires=[
        'bleach'
    ],
    test_suite='tests',
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from zenaura.client.component import Component, Reuseable
from zenaura.client.tags import Node 
from zenaura.client.compiler import compiler
from zenaura.client.hydrator import Hydrator

@Reuseable
class DefaultDomErrorComponent(Component):
    def __init__(self, error_message):
        super().__init__()
        self.error_message = error_message
    def render(self):
        return Node("div", children=[Node(text=str(self.error_message))])


class GracefulDegenerationLifeCycleWrapper(
    Hydrator
):
    
    def componentDidCatchError(self, comp, error) -> None:
        """
            Graceful degradation of component lifecycle methods.
        """
        # cleanup
        self.zen_dom_table.clear()

        if hasattr(comp, "componentDidCatchError"):
            # call componentDidCatchError method
            # mount the error message component 
            error_comp = comp.componentDidCatchError(str(error))
            compiled_comp =  self.hyd_comp_compile_render(
                 error_comp
            )
            
            # attach to real dom
            self.hyd_rdom_attach_to_root(compiled_comp)

            # update virtual dom
            self.hyd_vdom_update_with_new_render(comp, error_comp.render())

        else:
            # mount the default error message component
            error_comp  = DefaultDomErrorComponent(error_message=str(error))
            
            #compile the comp
            compiled_comp =  self.hyd_comp_compile_render(
                 error_comp
             )
            
            # attach to real dom
            self.hyd_rdom_attach_to_root(compiled_comp)

            # update virtual dom
            self.hyd_vdom_update_with_new_render(comp, error_comp.render())
